Attribute VB_Name = "CmboRmou"
Option Explicit
DefInt A-Z
' (C) Copyright 2001, Burton Systems Software, www.burtonsys.com.
' +=>keyflag<=-- "&(#)%n, Burton version %v, %d"
'   "&(#)CmboRmou.bas, Burton version 11, 22-Mar-01"

'Permission is granted to reuse this code, in whole or in part, as long
'as this copyright notice is retained.

'This module permits MRU comboboxes to be enhanced by the addition of
'support for clicking the right mouse button on an item in the dropdown
'list to remove that item from the MRU list.
'
'To use this module, simply add calls to SubClassCombo and UnSubClassCombo
'to your .frm files.  UnSubClassCombo must be called (with no parameters)
'from the Form_Unload subroutine.  SubClassCombo should be called from
'the _GotFocus subroutine for each MRU combobox.  If there is only one
'MRU combobox on a form, you may prefer to call SubClassCombo from the
'form's _Activate subroutine.  Or, if there is only one MRU combobox
'in the entire program, you can call SubClassCombo from the Form_Load
'subroutine.  Note: it doesn't hurt to call SubClassCombo multiple times.
'
'When calling SubClassCombo, just pass the combobox control as the only
'parameter.
'
'Calling SubClassCombo for a different combobox will automatically
'un-subclass the previous combobox.  Only one combobox is subclassed at
'a time.  That's okay because only one combobox can have focus at a time.
'
'However, there is one very subtle, minor advantage to calling SubClassCombo
'before the combobox gets focus: it permits the user to activate the
'dropdown list by right-clicking on the little arrow.  Otherwise, he
'must left-click the little arrow to drop down the list, before he can
'right-click on an item to be deleted from the MRU.  That's why you might
'want to call SubClassCombo from From_Activate or Form_Load.



'API Constants
Public Const GWL_WNDPROC = (-4)
'Public Const WM_CTLCOLOREDIT = &H133  '307
Public Const WM_CTLCOLORLISTBOX = &H134  '308
Public Const WM_LBUTTONDOWN = &H201  '513
Public Const WM_LBUTTONUP = &H202  '514
Public Const WM_RBUTTONDOWN = &H204  '516
Public Const WM_RBUTTONUP = &H205  '517
Public Const CB_SHOWDROPDOWN = &H14F
Public Const CB_GETDROPPEDSTATE = &H157

'API Declares
#If UNICODE Then
    Public Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongW" (ByVal hwnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Any) As Long
    Public Declare Function GetWindowLong Lib "user32" Alias "GetWindowLongW" (ByVal hwnd As Long, ByVal nIndex As Long) As Long
    Public Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcW" (ByVal lpPrevWndFunc As Long, ByVal hwnd As Long, ByVal msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
#Else
    Public Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" (ByVal hwnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Long) As Long
    Public Declare Function GetWindowLong Lib "user32" Alias "GetWindowLongA" (ByVal hwnd As Long, ByVal nIndex As Long) As Long
    Public Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcA" (ByVal lpPrevWndFunc As Any, ByVal hwnd As Long, ByVal msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
#End If
Public Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hwnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long
Public Declare Function MoveWindow Lib "user32" (ByVal hwnd As Long, ByVal x As Long, ByVal y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal bRepaint As Long) As Long

'Static variables used by this module:
Dim OldWndProc As Long  'where SubClassCombo saves old WndProc address for ComboBox
Dim SubClassedComboBox As Control  'saved input parameter to SubClassCombo, an MRU list's combo box
Dim SubClassedCboHand As Long  'SubClassedComboBox.hwnd
Dim OldCboListWndProc As Long  'where SubClassComboList saves old WndProc address for SubClassedComboBox.List
'Dim Combo_Edit_hwnd As Long  'hwnd for SubClassedComboBox's editable textbox subcontrol (discovered by monitoring WM_CTLCOLOREDIT messages sent to SubClassedComboBox)
Dim Combo_List_hwnd As Long  'hwnd for SubClassedComboBox.List (discovered by monitoring WM_CTLCOLORLISTBOX messages sent to SubClassedComboBox)
Dim Previous_Combo_List_hwnd As Long
Dim Previous_SubClassedCboHand As Long
'---------------------------

'General purpose On-Error displayer.  Prefix$ is normally the name of the
'function in which the On-Error routine is running.
Sub ShowError(prefix$)
  On Error Resume Next
   Dim svErr As Long
   Dim svError$
   Dim svErl As Long
   svErr = Err
   svErl = Erl
   svError$ = Error$
   If 50003 = svErr Then
      svError$ = "You have out-of-date versions of one or more Microsoft DLLs and/or OCXs."
   End If
   If (svErl = 0) Then
      'note: Erl is only non-zero if there are Basic line numbers in source code
      MsgBox "ERROR: #" & svErr & ", " & prefix$ & ", " & svError$
   Else
      MsgBox "ERROR: #" & svErr & ", " & prefix$ & ", line " & svErl & ", " & svError$
   End If
End Sub

'SubClass the combo control's Listbox so we can respond to and use its messages
Private Sub SubClassComboList(iHwnd_cl As Long)
    On Error GoTo Er
    If 0 = OldCboListWndProc Then
       'First, we need to store the address of the existing Message Handler
       OldCboListWndProc = GetWindowLong(iHwnd_cl, GWL_WNDPROC)
       'Now we can tell windows to forward all messages to our Message Handler
       SetWindowLong iHwnd_cl, GWL_WNDPROC, AddressOf SubClassComboList_WndMessage
       Previous_Combo_List_hwnd = iHwnd_cl
    Else
       'this should be impossible, I think
       MsgBox "ERR: attempt to re-subclass CboList, iHwnd_cl=0x" & Hex(iHwnd_cl) & ", saved WndProc=0x" & Hex(OldCboListWndProc)
    End If

    Exit Sub
Er:
    ShowError "SubClassComboList"
    Resume Next
End Sub

'SubClass the combo control so we can respond to and use its messages
Public Sub SubClassCombo(CboBox As Control)
    On Error GoTo Er
    Dim iHwnd As Long
    Set SubClassedComboBox = CboBox
    iHwnd = CboBox.hwnd
    If "RMOUDEM" <> App.EXEName Then
       'If running in the VB IDE, then App.EXEName is ""RMOUDEM"" -- don't hook messages when running in the IDE!
       'Note: this is dependent on the actual .EXE name differing from the .vbp name
       If SubClassedCboHand <> iHwnd Then
          UnSubClassCombo  'in case another combo box was subclassed
          Set SubClassedComboBox = CboBox  'set it again, because UnSubClassCombo set it to Nothing
          'First, we need to store the address of the existing Message Handler
          SubClassedCboHand = iHwnd
          OldWndProc = GetWindowLong(iHwnd, GWL_WNDPROC)
          'Now we can tell windows to forward all messages to our Message Handler
          SetWindowLong iHwnd, GWL_WNDPROC, AddressOf SubClassCombo_WndMessage
          If Previous_SubClassedCboHand = iHwnd Then
             'same ComboBox we handled last time...
             If Previous_Combo_List_hwnd <> 0 Then
                '...so its drop-down listbox also has the same handle as last time, so we can go ahead and subclass it immediately, rather than awaiting a WM_CTLCOLORLISTBOX message
                Combo_List_hwnd = Previous_Combo_List_hwnd
                SubClassComboList (Combo_List_hwnd)
             End If
          Else
             Previous_Combo_List_hwnd = 0
             Previous_SubClassedCboHand = iHwnd
          End If
       End If
    End If

    Exit Sub
Er:
    ShowError "SubClassCombo"
    Resume Next
End Sub

'Return control of the messages to ComboBox's ListBox back to windows.
Private Sub UnSubClassComboList(iHwnd_cl As Long)
    On Error GoTo Er
    'The IF tests & resetting to zero are so that we can call this multiple
    'times w/o breaking anything
    If 0 <> OldCboListWndProc Then
       If 0 = iHwnd_cl Then
          'It is okay to call this with iHwnd_cl=0 only if SubClassComboList() wasn't called.
          MsgBox "ERR: UnSubClassComboList(0)"
       Else
          Call SetWindowLong(iHwnd_cl, GWL_WNDPROC, OldCboListWndProc)
          OldCboListWndProc = 0
       End If
    End If

    Exit Sub
Er:
    ShowError "UnSubClassComboList"
    Resume Next
End Sub

'Return control of the messages back to windows.
Public Sub UnSubClassCombo()
    On Error GoTo Er
    'The IF tests & resetting to zero are so that we can call this multiple
    'times w/o breaking anything
    UnSubClassComboList (Combo_List_hwnd)
    If 0 <> OldWndProc Then
       Call SetWindowLong(SubClassedCboHand, GWL_WNDPROC, OldWndProc)
       OldWndProc = 0
       SubClassedCboHand = 0
    End If
    Set SubClassedComboBox = Nothing

    Exit Sub
Er:
    ShowError "UnSubClassCombo"
    Resume Next
End Sub

'Handle messages sent to ComboBox
Public Function SubClassCombo_WndMessage(ByVal hwnd As Long, ByVal msg As Long, ByVal wp As Long, ByVal lp As Long) As Long
    On Error GoTo Er
    Dim ListIsVisible As Long
    Dim rc As Long

'   If (WM_CTLCOLOREDIT = msg) And (Combo_Edit_hwnd <> lp) Then
'      Combo_Edit_hwnd = lp
'   End If
    If (WM_CTLCOLORLISTBOX = msg) And ((Combo_List_hwnd <> lp) Or (0 = OldCboListWndProc)) Then
       If 0 <> Combo_List_hwnd Then
          UnSubClassComboList (Combo_List_hwnd)
       End If
       Combo_List_hwnd = lp
       SubClassComboList (Combo_List_hwnd)
    End If

    'Pass the message along so the combo draws etc.
    SubClassCombo_WndMessage = CallWindowProc(OldWndProc, hwnd, msg, wp, lp)

    'Make the right mouse button force the dropdown list to drop down, like the left mouse button does.
    'Note that when the cursor is over the editable text box portion of the combo box,
    'that sub-control intercepts the WM_RBUTTONDOWN message, so that we won't even see
    'it here.  Thus, this only makes the dropdown list drop down when they right-click
    'the little arrow at the right end.
    If WM_RBUTTONDOWN = msg Then
       ListIsVisible = SendMessage(hwnd, CB_GETDROPPEDSTATE, 0, 0)
       If 0 = ListIsVisible Then
          'Dropdown list isn't visible, so make it visible
          rc = SendMessage(hwnd, CB_SHOWDROPDOWN, 1, 0)
       End If
    End If

    Exit Function
Er:
    ShowError "SubClassCombo_WndMessage"
    Resume Next
End Function

'handle messages sent to ComboBox's ListBox
Public Function SubClassComboList_WndMessage(ByVal hwnd As Long, ByVal msg As Long, ByVal wp As Long, ByVal lp As Long) As Long
    On Error GoTo Er
    Dim t$
    Dim rc As VbMsgBoxResult
    Dim svListIndex As Integer

    'Pass the message along so the combo draws etc.
    SubClassComboList_WndMessage = CallWindowProc(OldCboListWndProc, hwnd, msg, wp, lp)

    'Deal with incoming messages
    If msg = WM_RBUTTONUP Then
       If SubClassedComboBox.ListIndex >= 0 Then
          svListIndex = SubClassedComboBox.ListIndex
          t$ = "Remove " & SubClassedComboBox.List(svListIndex) & " from MRU list?"
          rc = MsgBox(t$, vbYesNo, "Edit MRU list")
          If vbYes = rc Then
             SubClassedComboBox.RemoveItem (svListIndex)
          End If
       End If
    End If

    Exit Function
Er:
    ShowError "SubClassComboList_WndMessage"
    Resume Next
End Function

